#!/bin/sh

# Update the snapshot within a bzr branch using get-orig-source
# (c) 2007-2009, Fabien Tassin <fta@sofaraway.org>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

WHO=`bzr whoami`
DATE=`date -R`
NAME=`basename $0`

usage () {
  echo "Usage: $NAME [-n|--new] [-c conf] [-p pkg-conf] [-d|--date YYYYMMDDtHHMM] [-L .../branch] branch [branch2 [...]]"
  echo "       $NAME [-n|--new] [-D|--daily] [--test] [-t|--tag SOMETAG=someversion] branch"
  echo
  echo " By default, email, date and version are always updated."
  echo " To create a new version, use -n"
  echo " To prevent a bzr pull/push, use --test"
  echo " To prevent a bzr pull, and push in overwrite mode, use --fix"
  echo " To work from a local branch, use -L"
  echo " To update the package branch from the last daily snapshot, use --daily"
}

get_conf () {
  local conf=$1
  local key=$2
  perl -e '
    $/ = undef;
    open(F, "'$conf'") || print "Cant open conf file: $!\n";
    $s = <F>;
    close F;
    $data = eval $s;
    my @keys = split /\//, "'$key'";
    $d = $data;
    while (my $k = shift @keys) {
      exit 1 unless defined $$d{$k};
      $d = $$d{$k};
      if (ref($d) eq "ARRAY" && $#$keys == -1) {
        $d = join " ", @$d;
        last;
      }
    }
    print $d;
    exit 0;
  '
  return 0
}

set -e

want_new=0
want_test=0
want_fix=0
want_verbose=0
upgrade_from_daily=0
TAG=""
DATETAG=""
BRANCH=""
PKGCONF=""
while [ $# -gt 0 ]; do
  case "$1" in
    -t | --tag )
      shift
      if [ $# = 0 ] ; then
        echo Error: -t needs an argument
        exit 1
      fi
      TAG=DEBIAN_TAG=$1
      shift ;;
    -d | --date )
      shift
      if [ $# = 0 ] ; then
        echo Error: -t needs an argument
        exit 1
      fi
      DATETAG=DEBIAN_DATE=$1
      shift ;;
    -b | --branch )
      shift
      if [ $# = 0 ] ; then
        echo Error: -b needs an argument
        exit 1
      fi
      BRANCH=DEBIAN_BRANCH=$1
      shift ;;
    -n | --new )
      shift
      want_new=1 ;;
    -L | --local-branch )
      shift
      if [ $# = 0 ] ; then
        echo Error: -L needs an argument
        exit 1
      fi
      DIR=$1
      if [ ! -d "$DIR" ] ; then
        echo Error: directory $DIR does not exist
        exit
      fi
      DIR=`cd "$DIR"; pwd`
      LBRANCH=LOCAL_BRANCH="$DIR"
      shift ;;
    -c | --config )
      shift
      if [ $# = 0 ] ; then
        echo Error: -c needs an argument
        exit 1
      fi
      CONF=$1
      if [ ! -f "$CONF" ] ; then
        echo Error: file $CONF does not exist
        exit
      fi
      GCONF="$CONF"
      shift ;;
    -p | --pkgs-config )
      shift
      if [ $# = 0 ] ; then
        echo Error: -p needs an argument
        exit 1
      fi
      CONF=$1
      if [ ! -f "$CONF" ] ; then
        echo Error: file $CONF does not exist
        exit
      fi
      PKGCONF="$CONF"
      shift ;;
    -D | --daily )
      shift
      upgrade_from_daily=1 ;;
    -v | --verbose )
      shift
      want_verbose=1 ;;
    -h | --help )
      usage
      exit 0
      ;;
    --test )
      shift
      want_test=1 ;;
    --fix )
      shift
      want_fix=1 ;;
    -- ) # Stop option processing
      shift
      break ;;
    -* )
      echo Error: unknown option $1
      exit 1 ;;
    * )
      break ;;
  esac
done

if [ "Z$GCONF" != Z ] ; then
  W="`get_conf $GCONF who`"
  if [ "Z$W" != Z ] ; then
    WHO="$W"
  fi
fi

PKGS="$@"

if [ "Z${PKGS}Z" = "ZZ" ] ; then
  # Get the list of PPA branches to update from the pkg-conf 'order' field
  PROJECTS=`get_conf $PKGCONF order`
  SUFFIX=`get_conf $GCONF ppa_branchtag`
  for PROJECT in $PROJECTS ; do
    NO_AUTO=`get_conf $PKGCONF "pkgs/$PROJECT/no_auto"`
    if [ "Z$NO_AUTO" = Z1 ] || [ "Z$NO_AUTO" = Ztrue ] ; then
      continue
    fi
    NO_UPDATE=`get_conf $PKGCONF "pkgs/$PROJECT/no_update"`
    if [ "Z$NO_UPDATE" = Z1 ] || [ "Z$NO_UPDATE" = Ztrue ] ; then
      continue
    fi
    PPA=`get_conf $PKGCONF "pkgs/$PROJECT/ppa"`
    if [ "Z$PPA" = Z ] ; then
      PPA=`get_conf $PKGCONF "pkgs/$PROJECT/branch"`@$SUFFIX
    fi
    PKGS="$PKGS $PPA"
  done
fi

for PKG in $PKGS ; do
  SUFFIX=".`echo $PKG | grep @ | cut -d@ -f2`"
  if [ "Z$SUFFIX" != Z. ] ; then
    PKG=`echo $PKG | cut -d@ -f1`
  else
    SUFFIX=""
  fi
  echo "* Doing '$PKG'..."
  if [ ! -d $PKG ] ; then
    echo "'$PKG' is not a branch. Skip"
    continue
  fi

  if [ $upgrade_from_daily -eq 1 ] ; then
    # Update the packaging branch using just the last version & date
    # from the corresponding daily branch. This is useful when a
    # daily broke and the packaging needs a fix
    SPKG=`(cd $PKG ; head -1 debian/changelog | awk '{ print $1 }')`
    LP=`get_conf $PKGCONF "pkgs/$SPKG/lp"`
    SUFFIX=`get_conf $GCONF ppa_branchtag`
    DAILY=$PKG.$SUFFIX
    DVERSION=`(cd $DAILY ; dpkg-parsechangelog | grep Version: | cut -d' ' -f2-)`
    DDATE=`(cd $DAILY ; dpkg-parsechangelog | grep Date: | cut -d' ' -f2-)`
    DWHO=`(cd $DAILY ; dpkg-parsechangelog | grep Maintainer: | cut -d' ' -f2-)`
    V=`echo $DVERSION | sed -e 's/~cvs/ CVS /' -e 's/~svn/ SVN /' -e 's/~hg/ HG /' -e 's/~ / /' -e 's/+nobinonly//' -e 's/-.*//'`
    echo "Update $PKG using data found in $DAILY: $DVERSION, $DWHO, $DDATE"
    if [ $want_new = 1 ] ; then
      (cd $PKG ; DEBEMAIL="$DWHO" dch --no-conf --newversion $DVERSION --distribution UNRELEASED "New upstream snapshot: $V" )
    else
      sed -i -e "1s/(.*) /($DVERSION) /" $PKG/debian/changelog
      sed -i -e "1,/^ -- /s/^ -- .*/ -- $DWHO  $DDATE/" $PKG/debian/changelog
    fi
    cd $PKG
    bzr diff || test $? = 1
    bzr commit -m "* New upstream snapshot: $V"
    bzr push $LP
    cd ..
    continue
  fi

  GET_UPSTREAM=""
  DIR=""
  LBRANCH=""
  if [ "Z$PKGCONF" != Z ] ; then
    SPKG=`(cd $PKG ; head -1 debian/changelog | awk '{ print $1 }')`
    DIR=`get_conf $PKGCONF "pkgs/$SPKG/upstream" | sed -e 's,/*$,,'`
    if [ "Z$DIR" != Z ] ; then
      if [ ! -d "$DIR" ] ; then
        # The upstream branch does not exist yet, create it
        RDIR=`dirname $DIR`
        if [ ! -d "$RDIR" ] ; then
          mkdir -p "$RDIR"
        fi
        GET_UPSTREAM=`get_conf $PKGCONF "pkgs/$SPKG/local_init"`
        if [ "Z$GET_UPSTREAM" != Z ] ; then
          echo "- Getting upstream branch:\n$GET_UPSTREAM $DIR"
          $GET_UPSTREAM $DIR
          RET=$?
          echo RET=$RET
        fi
      else
        DIR=`cd "$DIR"; pwd`
      fi
      if [ ! -d "$DIR" ] ; then
        mkdir -p "$DIR"
        DIR=`cd "$DIR"; pwd`
      fi
      LBRANCH=LOCAL_BRANCH="$DIR"
    fi
  fi

  DPUT=`get_conf $GCONF dput_target`
  VERSION_PATTERN=`get_conf $PKGCONF "pkgs/$SPKG/vpattern"`
  LP=""
  LP_DEST=""
  if [ "Z$DPUT" != Z ] ; then
    ACCOUNT=`sed -e '0,/^\['$DPUT'\]$/d' < $HOME/.dput.cf | sed -e '/^\[/,$d' | grep -E '^[:blank:]*incoming' | \
       awk -F '=' '{ print $2 }' | cut -d/ -f1 | sed -e 's/^[ 	]*//'`
    if [ "Z`echo $ACCOUNT | grep '%('`" != Z ]; then
      echo "*FIXME* Unsupported dput syntax. See http://blog.launchpad.net/ppa/simplifying-dputcf-for-multiple-ppas"
      ACCOUNT=""
    fi
    if [ "Z$ACCOUNT" = Z ] ; then
      echo "*WARN* Unable to identify the user account for '$DPUT' from ~/.dput.cf. Please investigate."
    fi
    LP=`get_conf $PKGCONF "pkgs/$SPKG/lp"`
    if [ "Z$LP" != Z ] && [ "Z$ACCOUNT" != Z ] ; then
      LP_DEST=`echo $LP | sed -e 's/lp:[^/]*/lp:'$ACCOUNT'/' | sed -e 's,/[^/]*$,/'$PKG$SUFFIX','`
    fi
  fi
  cd $PKG
  if [ $want_test = 1 ] ; then
    echo "*INFO* bzr pull disabled during crash troubleshooting"
  else
    if [ $want_fix = 1 ] ; then
      echo "*INFO* bzr pull disabled"
    else
      bzr pull $LP || ( [ $? -eq 3 ] && bzr push $LP )
    fi
  fi

  echo -n debian/rules get-orig-source $TAG $DATETAG $LBRANCH $BRANCH...
  if [ $want_verbose -eq 1 ] ; then
    echo ""
    debian/rules get-orig-source $TAG $DATETAG $LBRANCH $BRANCH
    echo "# get-orig-source complete"
  else
    debian/rules get-orig-source $TAG $DATETAG $LBRANCH $BRANCH > /dev/null 2>&1
    echo " done"
  fi

  P1=`ls -rt *orig.tar.gz | tail -1`
  S1=`stat -c %s $P1`
  NAME=`echo $P1 | sed -e 's/_.*//'`
  VERSION=`echo $P1 | sed -e 's/.*_//' -e 's/\.orig\..*$//'`
  P2=`head -1 ../$PKG$SUFFIX/debian/changelog`
  ONAME=`echo $P2 | sed -e 's/ .*//'`
  OVERSION=`echo $P2 | sed -e 's/.*(\(.*\)) .*/\1/' -e 's/-0ubuntu.*//'`

  S2=0
  if [ -f "../tarballs/${ONAME}_${OVERSION}.orig.tar.gz" ] ; then
    S2=`stat -c %s "../tarballs/${ONAME}_${OVERSION}.orig.tar.gz"`
  fi
  echo Got $NAME $VERSION vs $ONAME $OVERSION
  if [ $NAME != $ONAME ] ; then
    echo "*ERROR* $NAME != $ONAME"
    exit
  fi
  if [ "Z$VERSION_PATTERN" != Z ] ; then
    if [ Z`echo $VERSION | grep -E "$VERSION_PATTERN"` != Z$VERSION ] ; then
      echo "*ERROR* $VERSION doesn't match $VERSION_PATTERN. Abort"
      rm $P1
      cd ..
      continue
    fi
  fi
  if [ $VERSION = $OVERSION ] ; then
    echo "No new src snapshot. Skip\n"
    rm $P1
    cd ..
    continue
  else
   echo -n "*INFO* new upstream snapshot: $NAME $OVERSION ($S2) -> $VERSION ($S1)"
   if [ $S2 -ne 0 ] ; then
     DELTA=`expr $S2 - $S1 | sed -e 's/-/_/'`
     printf " (%+dkB, %+.2f%%)" `echo "$DELTA 1000 / p " | dc` `echo "4 k $DELTA $S1 / 100 * p" | dc`
   fi
   echo
  fi

  cd ../$PKG$SUFFIX
  V=`echo $VERSION | sed -e 's/~cvs/ CVS /' -e 's/~svn/ SVN /' -e 's/~hg/ HG /' -e 's/~ / /' -e 's/+nobinonly//'`
  if [ $want_new = 1 ] ; then
    DEBEMAIL="$WHO" dch --no-conf --newversion $VERSION-0ubuntu1 --distribution UNRELEASED "New upstream snapshot: $V"
  else
    sed -i -e "1s/(.*) /($VERSION-0ubuntu1) /" debian/changelog
    sed -i -e "1,/^ -- /s/^ -- .*/ -- $WHO  $DATE/" debian/changelog
  fi
  bzr diff || test $? = 1
  bzr commit -m "* New upstream snapshot: $V"
  if [ ! -d ../tarballs ] ; then
    mkdir ../tarballs
  fi
  mv ../$PKG/$P1 ../tarballs/
  PUSH_ARGS=""
  if [ $want_fix = 1 ] ; then
    PUSH_ARGS=" --overwrite"
  fi
  if [ $want_test = 1 ] ; then
    echo "*INFO* bzr push disabled during crash troubleshooting"
  else
    bzr push $PUSH_ARGS $LP_DEST
  fi
  cd ..
done
